# -*- coding: utf-8 -*-
import arcpy
import os
from importlib import reload

import python.scripts_utilities.utils_functions
reload(python.scripts_utilities.utils_functions)
from python.scripts_utilities.utils_functions import * 

import python.scripts_utilities.importer_config
reload(python.scripts_utilities.importer_config)
from python.scripts_utilities.importer_config import *

import python.scripts_utilities.prioritization_config
reload(python.scripts_utilities.prioritization_config)
from python.scripts_utilities.prioritization_config import * 


class Identify_Crossing_Gaps(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Crossing Gap Analysis Tool"
        self.description = "Identify Street Corridors with long sections without suitable pedestrian crossings"
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""

        streets_fc = arcpy.Parameter(
            name="streets_fc",
            displayName="Streets Layer",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Input"
        )
        streets_fc.filter.list=["Polyline", "Multipolyline"]

        street_name_field = arcpy.Parameter(
            name="street_name_field",
            displayName="Street Name Field",
            datatype="Field",
            parameterType="Required",
            direction="Input"
        )
        street_name_field.parameterDependencies = [streets_fc.name]
        street_name_field.filter.list=["TEXT"]

        fclass_field = arcpy.Parameter(
            name="fclass_field",
            displayName="Functional Class Field",
            datatype="Field",
            parameterType="Required",
            direction="Input"
        )
        fclass_field.parameterDependencies = [streets_fc.name]
        fclass_field.filter.list=["TEXT", "SHORT", "LONG"]

        streets_filter = arcpy.Parameter(
            name="streets_filter",
            displayName="Street filter to apply (uses all streets if left blank)",
            datatype="GPSQLExpression",
            parameterType="Optional",
            direction="Input"
        )
        streets_filter.parameterDependencies=[streets_fc.name]

        gdb_dir = arcpy.Parameter(
            name="gdb_dir",
            displayName="Folder with Standardized and Scratch Geodatabases",
            datatype="DEFolder",
            parameterType="Required",
            direction="Input"
        )

        crossing_gap_length = arcpy.Parameter(
            name="crossing_gap_length",
            displayName="Maximum gap between crossings (in feet)",
            datatype="GPDouble",
            parameterType="Required",
            direction="Input"
        )
        crossing_gap_length.value = 1320

        output_fc = arcpy.Parameter(
            name="output_fc",
            displayName="Output Crossing Gaps Layer",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Output"
        )
        output_fc.value = os.path.join(arcpy.env.workspace, "Crossing_Gaps_Output")
        output_fc.symbology = os.path.join(symbology_path, "Crossing_Gaps_Output.lyrx")

        low_stress_crossing_locations = arcpy.Parameter(
            name="low_stress_crossing_locations",
            displayName="Low-Stress Crossing Locations",
            datatype="GPFeatureLayer",
            parameterType="Optional",
            direction="Input"
        )
        low_stress_crossing_locations.filter.list=["Point"]

        signals_are_low_stress = arcpy.Parameter(
            name="signals_are_low_stress",
            displayName="Are signals to be modeled as low-stress crossings? (in addition to the above low-stress crossing layer)",
            datatype="GPBoolean",
            parameterType="Required",
            enabled=True,
            direction="Input"
        )
        signals_are_low_stress.value = True

        high_stress_crossing_locations = arcpy.Parameter(
            name="high_stress_crossing_locations",
            displayName="High-Stress Crossing Locations",
            datatype="GPFeatureLayer",
            parameterType="Optional",
            direction="Input"
        )
        high_stress_crossing_locations.filter.list=["Point"]

        params = [
            streets_fc,
            street_name_field,
            fclass_field,
            streets_filter,
            gdb_dir,
            crossing_gap_length,
            output_fc,
            low_stress_crossing_locations,
            signals_are_low_stress,
            high_stress_crossing_locations
            ]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        # read parameters
        streets_fc = _get_by_name(parameters, "streets_fc").valueAsText
        street_name_field = _get_by_name(parameters, "street_name_field").valueAsText
        fclass_field = _get_by_name(parameters, "fclass_field").valueAsText
        streets_filter = _get_by_name(parameters, "streets_filter").valueAsText
        gdb_dir = _get_by_name(parameters, "gdb_dir").valueAsText
        crossing_gap_length = _get_by_name(parameters, "crossing_gap_length").value
        output_fc = _get_by_name(parameters, "output_fc").valueAsText
        low_stress_crossing_locations = _get_by_name(parameters, "low_stress_crossing_locations").valueAsText
        signals_are_low_stress = _get_by_name(parameters, "signals_are_low_stress").value
        high_stress_crossing_locations = _get_by_name(parameters, "high_stress_crossing_locations").valueAsText

        # ensure at least one of signals or low-stress crossings are provided
        # if not low_stress_crossing_locations and not signals_are_low_stress:
        #     arcpy.AddError("You must provide low-stress crossings layer or use signals layer.")
        #     raise arcpy.ExecuteError

        # ensure the proper gdb exists in the location specified
        standard_gdb = os.path.join(gdb_dir, "Standardized_BMC_Ped_Infra_Assessment.gdb")
        scratch_gdb = os.path.join(gdb_dir, "Scratch_BMC_Ped_Infra_Assessment.gdb")

        # Set environment settings
        arcpy.env.workspace = gdb_dir
        workspaces = arcpy.ListWorkspaces("*", "FileGDB")
        arcpy.env.overwriteOutput = True

        if not bool(set(workspaces) == set([standard_gdb, scratch_gdb])):
            arcpy.AddError("Geodatabase names are not valid. This can result from A) incorrect geodatabase names, B) missing geodatabases, C) too many geodatabases in folder. Please refer to the manual for the required geodatabases.")
            raise arcpy.ExecuteError

        arcpy.AddMessage("Creating street corridors based on street name and functional classification...")
        # copy streets to scratch gdb
        tmp_streets_fc = os.path.join(scratch_gdb, "tmp_streets_fc")
        fields_for_dissolve = [street_name_field, fclass_field]
        if streets_filter is None or streets_filter == "":
            copy_fc_with_fields(in_fc=streets_fc, out_fc=tmp_streets_fc, fields=fields_for_dissolve, scratch_gdb=scratch_gdb)
        else:
            copy_fc_with_fields(in_fc=streets_fc, out_fc=tmp_streets_fc, fields=fields_for_dissolve, scratch_gdb=scratch_gdb, where_clause=streets_filter)
        # clean up street names for those that start with N, S, E, W for dissolving
        arcpy.management.AddField(tmp_streets_fc, f"{street_name_field}_clean", 'TEXT')
        cursor = arcpy.da.UpdateCursor(tmp_streets_fc, field_names=[street_name_field, f"{street_name_field}_clean"])
        for row in cursor:
            if row[0][:2] in ['W ', 'E ', 'S ', 'N ']:
                row[1] = row[0][2:len(row[0])]
            else:
                row[1] = row[0]
            cursor.updateRow(row)
        del cursor
        
        # dissolve them by name and fclass
        tmp_streets_dissolved = os.path.join(scratch_gdb, "tmp_streets_dissolved")
        fields_for_dissolve = [f"{street_name_field}_clean", fclass_field]
        arcpy.Dissolve_management(tmp_streets_fc, tmp_streets_dissolved, dissolve_field=fields_for_dissolve)

        arcpy.AddMessage("Reading low stress and high stress crossing locations...")
        # get low stress crossing locations
        signals_fc = os.path.join(standard_gdb, getLayerName(layerInfo, "Traffic Signals"))
        tmp_low_stress_crossings = os.path.join(scratch_gdb, "tmp_low_stress_crossings")
        if low_stress_crossing_locations and signals_are_low_stress:
            tmp_extra_low_stress_crossings = os.path.join(scratch_gdb, "tmp_extra_low_stress_crossings")
            copy_fc_with_fields(
                in_fc=low_stress_crossing_locations,
                out_fc=tmp_extra_low_stress_crossings,
                fields=[],
                scratch_gdb=scratch_gdb
            )
            # to_merge.append(tmp_low_stress_crossings)
            arcpy.Merge_management([tmp_extra_low_stress_crossings, signals_fc], tmp_low_stress_crossings)
        elif low_stress_crossing_locations:
            copy_fc_with_fields(
                in_fc=low_stress_crossing_locations,
                out_fc=tmp_low_stress_crossings,
                fields=[],
                scratch_gdb=scratch_gdb
            )
        else:
            arcpy.CopyFeatures_management(signals_fc, tmp_low_stress_crossings)

        # remove anything that is wihtin a 30 feet of a high stress crossing
        if high_stress_crossing_locations:
            tmp_high_stress_locations = os.path.join(scratch_gdb, "tmp_high_stress_locations")
            copy_fc_with_fields(
                in_fc=high_stress_crossing_locations,
                out_fc=tmp_high_stress_locations,
                fields=[],
                scratch_gdb=scratch_gdb
            )
            # calculate field to flag anything that should be excluded
            select_layer = arcpy.SelectLayerByLocation_management(tmp_low_stress_crossings, "WITHIN_A_DISTANCE", tmp_high_stress_locations, "30 Feet", "NEW_SELECTION", "NOT_INVERT")
            arcpy.CalculateField_management(select_layer, "is_near_high_stress", "1", "PYTHON3", field_type="SHORT")
            arcpy.CalculateField_management(tmp_low_stress_crossings, "is_near_high_stress", "int(bool(!is_near_high_stress!))", "PYTHON3")
            cursor = arcpy.da.UpdateCursor(tmp_low_stress_crossings, ["is_near_high_stress"], where_clause="is_near_high_stress = 1")
            for row in cursor:
                cursor.deleteRow()
            del cursor
        
        arcpy.AddMessage("Breaking streets based on location of low-stress crossings...")
        tmp_low_stress_30ft_buffer = os.path.join(scratch_gdb, "tmp_low_stress_30ft_buffer")
        arcpy.Buffer_analysis(tmp_low_stress_crossings, tmp_low_stress_30ft_buffer, "30 Feet", dissolve_option="ALL")

        # now erase the low stress buffer from the dissolved streets
        tmp_streets_clipped = os.path.join(scratch_gdb, "tmp_streets_clipped")
        arcpy.PairwiseErase_analysis(tmp_streets_dissolved, tmp_low_stress_30ft_buffer, tmp_streets_clipped)

        tmp_crossing_gaps = os.path.join(scratch_gdb, "tmp_crossing_gaps")
        arcpy.MultipartToSinglepart_management(tmp_streets_clipped, tmp_crossing_gaps)
        arcpy.AddField_management(tmp_crossing_gaps, "crossing_gap", "TEXT")
        cursor = arcpy.da.UpdateCursor(tmp_crossing_gaps, ["SHAPE@LENGTH", "crossing_gap"])
        for row in cursor:
            if row[0] > crossing_gap_length:
                row[1] = "TRUE"
            else:
                row[1] = "FALSE"
            cursor.updateRow(row)
        del cursor
        
        # copy to output
        arcpy.management.CopyFeatures(tmp_crossing_gaps, output_fc)
        return True


def _get_by_name(param_list, name):
    list = [param_list[i].name for i in range(len(param_list))]
    return param_list[list.index(name)]