# -*- coding: utf-8 -*-
import arcpy
import os
from importlib import reload

import python.scripts_utilities.utils_functions
reload(python.scripts_utilities.utils_functions)
from python.scripts_utilities.utils_functions import * 

import python.scripts_utilities.importer_config
reload(python.scripts_utilities.importer_config)
from python.scripts_utilities.importer_config import *

import python.scripts_utilities.prioritization_config
reload(python.scripts_utilities.prioritization_config)
from python.scripts_utilities.prioritization_config import * 


class Identify_Sidewalk_Gaps(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Sidewalk Gap Analysis Tool"
        self.description = "Idenitfies segments on a street network that have gaps in sidewalks on left, right, or both sides of the street."
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""

        street_centerlines = arcpy.Parameter(
            name="street_centerlines",
            displayName="Street Centerline Layer",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Input"
        )
        street_centerlines.filter.list=["POLYLINE", "MULTIPOLYLINE"]

        primary_key = arcpy.Parameter(
            name="primary_key",
            displayName="Primary Key",
            datatype="GPString",
            parameterType="Required",
            direction="Input"
        ) 
        primary_key.filter.type = "ValueList"
        primary_key.filter.list = ["Select a Layer of Street Centerline Features"]

        buffer_value_type = arcpy.Parameter(
            name="buffer_value_type",
            displayName="Buffer Value Type",
            datatype="GPString",
            enabled=True,
            parameterType="Required",
            direction="Input"
        ) 
        buffer_value_type.filter.type = "ValueList"
        buffer_value_type.filter.list = ["Value", "Field"]
        buffer_value_type.value = "Value"

        buffer_distance = arcpy.Parameter(
            name="buffer_distance",
            displayName="Buffer Distance (in feet).",
            datatype="GPLong",
            parameterType="Optional",
            enabled=False,
            direction="Input"
        )
        buffer_distance.value = 50

        buffer_distance_field = arcpy.Parameter(
            name="buffer_distance_field",
            displayName=f"Select Buffer Distance Field (must be interger value in feet)",
            datatype="Field",
            parameterType="Optional",
            enabled=False,
            direction="Input"
        )
        buffer_distance_field.parameterDependencies = [street_centerlines.name]
        buffer_distance_field.filter.list=["LONG", "SHORT"]

        divided_street_query = arcpy.Parameter(
            name="divided_street_query",
            displayName="Query to identify each side of a divided road.",
            datatype="GPSQLExpression",
            parameterType="Optional",
            direction="Input"
        )
        divided_street_query.parameterDependencies=[street_centerlines.name]

        gdb_dir = arcpy.Parameter(
            name="gdb_dir",
            displayName="Folder with Standardized and Scratch Geodatabases",
            datatype="DEFolder",
            parameterType="Required",
            enabled=True,
            direction="Input"
        )

        output_fc = arcpy.Parameter(
            name="output_fc",
            displayName="Output Feature Class with Scores",
            datatype="GPFeatureLayer",
            parameterType="Required",
            enabled=True,
            direction="Output"
        )
        output_fc.value = os.path.join(arcpy.env.workspace, "Sidewalk_Gaps_Output")
        output_fc.symbology = os.path.join(symbology_path, "Sidewalk_Gaps_Output.lyrx")

        sidewalks_fc = arcpy.Parameter(
            name="sidewalks_fc",
            displayName="Sidewalks layer (if not provided, uses layer in standardized GDB)",
            datatype="GPFeatureLayer",
            parameterType="Optional",
            enabled=True,
            direction="Input"
        )
        sidewalks_fc.filter.list=["POLYLINE", "MULTIPOLYLINE"]

        max_sw_shortage = arcpy.Parameter(
            name="max_sw_shortage",
            displayName="Sidewalk deficit length in feet (Street is considered to have a gap if sidewalk length is shorter than street length by this value)",
            datatype="GPDouble",
            parameterType="Required",
            enabled=True,
            direction="Input"
        )
        max_sw_shortage.value=100

        params = [
            street_centerlines,
            primary_key,
            buffer_value_type,
            buffer_distance,
            buffer_distance_field,
            divided_street_query,
            gdb_dir,
            output_fc,
            sidewalks_fc,
            max_sw_shortage
            ]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""


        if parameters[0].valueAsText != None and parameters[0].enabled: 
            fc_attributes =[f.name for f in arcpy.ListFields(parameters[0].valueAsText) if not f.required] 
            parameters[1].filter.list = fc_attributes
        
        if parameters[2].valueAsText == "Value":
            parameters[3].enabled=True
            parameters[3].parameterType="Required"
            parameters[4].enabled=False
            parameters[4].parameterType="Optional"
        elif parameters[2].valueAsText == "Field":
            parameters[4].enabled=True
            parameters[4].parameterType="Required"
            parameters[4].parameterDependencies = [parameters[0].name]
            parameters[4].filter.list=["LONG", "SHORT"]
            parameters[3].enabled=False
            parameters[3].parameterType="Optional"

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        street_centerlines = _get_by_name(parameters, "street_centerlines").valueAsText
        primary_key = _get_by_name(parameters, "primary_key").valueAsText
        buffer_value_type = _get_by_name(parameters, "buffer_value_type").valueAsText
        max_sw_shortage = _get_by_name(parameters, "max_sw_shortage").value

        # parse divided street query
        divided_st_dict = defaultdict(int)
        if _get_by_name(parameters, "divided_street_query").value:
            divided_street_query = _get_by_name(parameters, "divided_street_query").valueAsText
            cursor = arcpy.da.SearchCursor(street_centerlines, [primary_key], where_clause = divided_street_query)
            for row in cursor:
                divided_st_dict[row[0]] = 1
            del cursor
        # else:
        #     divided_street_query = False
        
        # make sure a buffer values is provided
        if bool(_get_by_name(parameters, "buffer_distance").value) + bool(_get_by_name(parameters, "buffer_distance_field").value) == 0:
            arcpy.AddError("A buffer field or distance must be is provided.")
            raise arcpy.ExecuteError

        buffer_dist_dict = defaultdict(int)
        if buffer_value_type == "Value":
            cursor = arcpy.da.SearchCursor(street_centerlines, [primary_key])
            for row in cursor:
                buffer_dist_dict[row[0]] = f"{_get_by_name(parameters, 'buffer_distance').value} Feet"
            del cursor
        elif buffer_value_type == "Field":
            cursor = arcpy.da.SearchCursor(street_centerlines, [primary_key, _get_by_name(parameters, "buffer_distance_field").valueAsText])
            for row in cursor:
                if row[1] > 0:
                    buffer_dist_dict[row[0]] = f"{row[1]} Feet"
                else:
                    buffer_dist_dict[row[0]] = f"0.1 Feet"   # adding a small buffer as I am not sure if 0 buffer distance will cause issues
            del cursor

        
        gdb_dir = _get_by_name(parameters, "gdb_dir").valueAsText
        output_fc = _get_by_name(parameters, "output_fc").valueAsText

        # ensure the proper gdb exists in the location specified
        standard_gdb = os.path.join(gdb_dir, "Standardized_BMC_Ped_Infra_Assessment.gdb")
        scratch_gdb = os.path.join(gdb_dir, "Scratch_BMC_Ped_Infra_Assessment.gdb")

        # Set environment settings
        arcpy.env.workspace = gdb_dir
        workspaces = arcpy.ListWorkspaces("*", "FileGDB")
        arcpy.env.overwriteOutput = True

        if not bool(set(workspaces) == set([standard_gdb, scratch_gdb])):
            arcpy.AddError("Geodatabase names are not valid. This can result from A) incorrect geodatabase names, B) missing geodatabases, C) too many geodatabases in folder. Please refer to the manual for the required geodatabases.")
            raise arcpy.ExecuteError

        # make sure sidewalk layer is provided if the standard gdb does not have the layer
        if _get_by_name(parameters, "sidewalks_fc").value:
            sidewalks_fc = _get_by_name(parameters, "sidewalks_fc").valueAsText
            sw_filter = "1 = 1"
        elif arcpy.Exists(os.path.join(standard_gdb, getLayerName(layerInfo, "Pedestrian Network"))):
            sidewalks_fc = os.path.join(standard_gdb, getLayerName(layerInfo, "Pedestrian Network"))
            sw_filter = "CateTitle IN ('sidewalk')"
        else:
            arcpy.AddError("The standardized GDB does not have a 'Pedestrian Network' layer. Please upload a 'Pedestrian Network' layer to the standaized GDB or supply a sidewalk layer in the tool.")
            raise arcpy.ExecuteError
        
        # check if pkey field is unique
        # TODO check if this is necessary?
        if not is_field_unique(fc=street_centerlines, field=primary_key):
            arcpy.AddError("The primary key field supplied is either missing value or is being used by more than one features.")
            raise arcpy.ExecuteError
        
        # create a copy of the streets layer in the scratch database
        arcpy.AddMessage("Copying street centerlines to scratch workspace for analysis...")
        tmp_street_centerlines = os.path.join(scratch_gdb, "tmp_street_centerlines")
        fields =  [primary_key]
        rename_dict = {primary_key: "uid"}
        copy_fc_with_fields(street_centerlines, tmp_street_centerlines, fields=fields, scratch_gdb=scratch_gdb, rename_fields=rename_dict)
        
        # update buffer size
        arcpy.AddField_management(tmp_street_centerlines, "buffer_size", "TEXT")
        cursor = arcpy.da.UpdateCursor(tmp_street_centerlines, ["uid", "buffer_size"])
        for row in cursor:
            row[1] = buffer_dist_dict[row[0]]
            cursor.updateRow(row)
        del cursor

        # update divided street info
        arcpy.AddField_management(tmp_street_centerlines, "divided", "SHORT")
        cursor = arcpy.da.UpdateCursor(tmp_street_centerlines, ["uid", "divided"])
        for row in cursor:
            row[1] = divided_st_dict[row[0]]
            cursor.updateRow(row)
        del cursor

        tmp_sidewalks_fc = os.path.join(scratch_gdb, "tmp_sidewalks_fc")
        copy_fc_with_fields(sidewalks_fc, tmp_sidewalks_fc, fields=[], where_clause=sw_filter ,scratch_gdb=scratch_gdb)
        # note sidewalk length to be used in pariwise intersect later
        arcpy.CalculateField_management(tmp_sidewalks_fc, "uid_sw", "!OBJECTID!", "PYTHON3", field_type="LONG")
        sw_length_dict = dict()
        cursor = arcpy.da.SearchCursor(tmp_sidewalks_fc, ["uid_sw", "SHAPE@LENGTH"])
        for row in cursor:
            sw_length_dict[row[0]] = row[1]
        del cursor

        # create left and right buffers of the street centerlines layer
        if arcpy.CheckProduct("ArcInfo") in ("Available", "AlreadyInitialized"):
            for side in ["LEFT", "RIGHT"]:
                arcpy.AddMessage(f"Buffering for {side} of streets centerlines and calculating sidewalk overlap...")
                tmp_buffer = os.path.join(scratch_gdb, f"tmp_{side}_buffer")
                arcpy.Buffer_analysis(tmp_street_centerlines, tmp_buffer, "buffer_size", line_side=side, line_end_type="FLAT")
                tmp_buffer_sw_intersect = os.path.join(scratch_gdb, f"tmp_buffer_sw_intersect_{side}")
                arcpy.PairwiseIntersect_analysis([tmp_buffer, tmp_sidewalks_fc], tmp_buffer_sw_intersect)
                sw_intersect_length = defaultdict(int)
                cursor = arcpy.da.SearchCursor(tmp_buffer_sw_intersect, ["uid", "SHAPE@LENGTH", "uid_sw", "buffer_size"])
                for row in cursor:
                    buffer_in_ft = int(row[3][:-5])
                    if row[1] > 1.5*buffer_in_ft or sw_length_dict[row[2]] <= 1.5*buffer_in_ft:
                        sw_intersect_length[row[0]] += row[1]
                del cursor
                
                # write the intersect distance to the streets file
                arcpy.CalculateField_management(tmp_street_centerlines, f"sw_length_{side}", "0", "PYTHON3", field_type="DOUBLE")
                arcpy.CalculateField_management(tmp_street_centerlines, f"sw_gap_length_{side}", "0", "PYTHON3", field_type="DOUBLE")
                arcpy.AddField_management(tmp_street_centerlines, f"sw_gap_{side}", "TEXT")
                cursor = arcpy.da.UpdateCursor(tmp_street_centerlines, ["uid", f"sw_length_{side}", f"sw_gap_{side}", "SHAPE@LENGTH", f"sw_gap_length_{side}"])
                for row in cursor:
                    row[1] = sw_intersect_length[row[0]]
                    row[4] = row[3] - sw_intersect_length[row[0]]
                    if row[3] - sw_intersect_length[row[0]] >= max_sw_shortage:
                        row[2] = 'True'
                    else:
                        row[2] = 'False'
                    cursor.updateRow(row)
                del cursor
            # now add a sidewalk gap type field
            arcpy.AddField_management(tmp_street_centerlines, "sw_gap_type", "TEXT")
            cursor = arcpy.da.UpdateCursor(tmp_street_centerlines, ["sw_gap_LEFT", "sw_gap_RIGHT", "divided", "sw_gap_type"])
            for row in cursor:
                (gap_left, gap_right, divided, gap_type) = row
                if gap_left == 'False' and gap_right == 'False':
                    row[3] = 'No Gap'
                elif (gap_left == 'True' and gap_right == 'False') or (gap_left == 'False' and gap_right == 'True'):
                    if divided == 1:
                        row[3] = 'No Gap'
                    else:
                        row[3] = 'Gap on One Side'
                elif gap_left == 'True' and gap_right == 'True':
                    if divided == 1:
                        row[3] = 'Gap on Divided Roadway'
                    else:
                        row[3] = 'Gap on Both Sides'
                cursor.updateRow(row)
            del cursor
        else:
            arcpy.AddWarning("Advanced license is not available. Using a double sided buffer instead of each side separately.")
            tmp_buffer = os.path.join(scratch_gdb, f"tmp_both_sides_buffer")
            arcpy.Buffer_analysis(tmp_street_centerlines, tmp_buffer, "buffer_size")
            tmp_buffer_sw_intersect = os.path.join(scratch_gdb, f"tmp_buffer_sw_intersect_both_sides")
            arcpy.PairwiseIntersect_analysis([tmp_buffer, tmp_sidewalks_fc], tmp_buffer_sw_intersect)
            sw_intersect_length = defaultdict(int)
            cursor = arcpy.da.SearchCursor(tmp_buffer_sw_intersect, ["uid", "SHAPE@LENGTH", "uid_sw", "buffer_size"])
            for row in cursor:
                buffer_in_ft = int(row[3][:-5])
                if row[1] > 1.5*buffer_in_ft or sw_length_dict[row[2]] <= 1.5*buffer_in_ft:
                    sw_intersect_length[row[0]] += row[1]
            del cursor
            
            # write the intersect distance to the streets file
            arcpy.CalculateField_management(tmp_street_centerlines, f"sw_length_both_sides", "0", "PYTHON3", field_type="DOUBLE")
            arcpy.CalculateField_management(tmp_street_centerlines, f"sw_gap_length_both_sides", "0", "PYTHON3", field_type="DOUBLE")
            arcpy.AddField_management(tmp_street_centerlines, f"sw_gap_both_sides", "TEXT")
            arcpy.AddField_management(tmp_street_centerlines, "sw_gap_type", "TEXT")
            cursor = arcpy.da.UpdateCursor(tmp_street_centerlines, ["uid", f"sw_length_both_sides", f"sw_gap_both_sides", "SHAPE@LENGTH", "divided", f"sw_gap_length_both_sides", "sw_gap_type"])
            for row in cursor:
                row[1] = sw_intersect_length[row[0]]
                row[5] = row[3] - sw_intersect_length[row[0]]
                if row[4] == 1:
                    if row[3] - sw_intersect_length[row[0]] >= max_sw_shortage:
                        row[2] = 'True'
                        row[6] = 'Gap on Divided Roadway'
                    else:
                        row[2] = 'False'
                        row[6] = 'No Gap'
                else:
                    if row[3] - (sw_intersect_length[row[0]] / 2) >= max_sw_shortage:
                        row[2] = 'True'
                        row[6] = 'Gap on One or Both Sides'
                    else:
                        row[2] = 'False'
                        row[6] = 'No Gap'
                cursor.updateRow(row)
            del cursor
        
        arcpy.management.CopyFeatures(tmp_street_centerlines, output_fc)
        return True


def _get_by_name(param_list, name):
    list = [param_list[i].name for i in range(len(param_list))]
    return param_list[list.index(name)]