# -*- coding: utf-8 -*-
from xml.etree.ElementInclude import include
import arcpy
import os
import shutil
import pandas as pd
from openpyxl import load_workbook
from importlib import reload

import python.scripts_utilities.utils_functions
reload(python.scripts_utilities.utils_functions)
from python.scripts_utilities.utils_functions import * 

import python.scripts_utilities.importer_config
reload(python.scripts_utilities.importer_config)
from python.scripts_utilities.importer_config import *

import python.scripts_utilities.prioritization_config
reload(python.scripts_utilities.prioritization_config)
from python.scripts_utilities.prioritization_config import * 


class Travelshed_Analysis(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Pedestrian Travelshed Analysis Tool"
        self.description = "Identify pedestrian travelshed from a given location using different networks."
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""

        destination_fc = arcpy.Parameter(
            name="destination_fc",
            displayName="Destination Layer",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Input"
        )

        buffer_distance = arcpy.Parameter(
            name="buffer_distance",
            displayName="Buffer distance to apply around destinations for routing",
            datatype="GPLinearUnit",
            parameterType="Required",
            direction="Input"
        )
        buffer_distance.value = "50 Feet"

        travel_distance = arcpy.Parameter(
            name="travel_distance",
            displayName="Travel distance (in miles)",
            datatype="GPDouble",
            parameterType="Required",
            direction="Input"
        )
        travel_distance.value = "0.5"

        gdb_dir = arcpy.Parameter(
            name="gdb_dir",
            displayName="Folder with Standardized and Scratch Geodatabases",
            datatype="DEFolder",
            parameterType="Required",
            direction="Input"
        )

        optional_routing_layer = arcpy.Parameter(
            name="optional_routing_layer",
            displayName="Optional layer to run routing for travelshed analysis",
            datatype="GPFeatureLayer",
            parameterType="Optional",
            direction="Input"
        )

        append_to_sw_network = arcpy.Parameter(
            name="append_to_sw_network",
            displayName="Should the 'Pedestrian Network' layer in standard GDB be added to this layer for routing?",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=False,
            direction="Input"
        )
        append_to_sw_network.value = True

        output_euclidean_reach_fc = arcpy.Parameter(
            name="output_euclidean_reach_fc",
            displayName="Output feature class for the travelshed as the crow flies",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Output"
        )
        output_euclidean_reach_fc.value = os.path.join(arcpy.env.workspace, "Euclidean_Travelshed")

        output_sw_reach_fc = arcpy.Parameter(
            name="output_sw_reach_fc",
            displayName="Output feature class for travelshed using sidewalk network",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Output"
        )
        output_sw_reach_fc.value = os.path.join(arcpy.env.workspace, "Sidewalk_Network_Travelshed")

        output_optional_reach_fc = arcpy.Parameter(
            name="output_optional_reach_fc",
            displayName="Output feature class for travelshed using optional layer provided",
            datatype="GPFeatureLayer",
            parameterType="Required",
            enabled=False,
            direction="Output"
        )
        output_optional_reach_fc.value = os.path.join(arcpy.env.workspace, "Optional_Network_Travelshed")

        output_sw_reach_fc_buffer = arcpy.Parameter(
            name="output_sw_reach_fc_buffer",
            displayName="Output feature class (buffered) for travelshed using sidewalk network",
            datatype="GPFeatureLayer",
            parameterType="Required",
            enabled=False,
            direction="Output"
        )

        output_optional_reach_fc_buffer = arcpy.Parameter(
            name="output_optional_reach_fc_buffer",
            displayName="Output feature class (buffered) for travelshed using optional layer provided",
            datatype="GPFeatureLayer",
            parameterType="Required",
            enabled=False,
            direction="Output"
        )
        output_optional_reach_fc.value = os.path.join(arcpy.env.workspace, "Optional_Network_Travelshed")
        # output_optional_reach_fc.value = "Select a valid layer"

        params = [
            destination_fc,
            buffer_distance,
            travel_distance,
            gdb_dir,
            optional_routing_layer,
            append_to_sw_network,
            output_euclidean_reach_fc,
            output_sw_reach_fc,
            output_optional_reach_fc,
            output_sw_reach_fc_buffer,
            output_optional_reach_fc_buffer
            ]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""

        optional_routing_layer = _get_by_name(parameters, "optional_routing_layer")
        if optional_routing_layer.value:
            _get_by_name(parameters, "append_to_sw_network").enabled = True
            _get_by_name(parameters, "output_optional_reach_fc").enabled = True
            # if _get_by_name(parameters, "output_optional_reach_fc").valueAsText == "placeholder":
            #     _get_by_name(parameters, "output_optional_reach_fc").value = ""
        else:
            _get_by_name(parameters, "append_to_sw_network").enabled = False
            _get_by_name(parameters, "output_optional_reach_fc").enabled = False
            # _get_by_name(parameters, "output_optional_reach_fc").value = "placeholder"
        
        # update buffered layer output values
        output_sw_reach_fc = _get_by_name(parameters, "output_sw_reach_fc")
        if output_sw_reach_fc.value:
            _get_by_name(parameters, "output_sw_reach_fc_buffer").value = str(output_sw_reach_fc.value) + "_buffer"
        
        output_optional_reach_fc = _get_by_name(parameters, "output_optional_reach_fc")
        if output_optional_reach_fc.value:
            _get_by_name(parameters, "output_optional_reach_fc_buffer").value = str(output_optional_reach_fc.value) + "_buffer"

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        destination_fc = _get_by_name(parameters, "destination_fc").valueAsText
        buffer_distance = _get_by_name(parameters, "buffer_distance").valueAsText
        travel_distance = _get_by_name(parameters, "travel_distance").value
        gdb_dir = _get_by_name(parameters, "gdb_dir").valueAsText
        optional_routing_layer = _get_by_name(parameters, "optional_routing_layer").valueAsText
        append_to_sw_network = _get_by_name(parameters, "append_to_sw_network").value
        output_euclidean_reach_fc = _get_by_name(parameters, "output_euclidean_reach_fc").valueAsText
        output_sw_reach_fc = _get_by_name(parameters, "output_sw_reach_fc").valueAsText
        output_optional_reach_fc = _get_by_name(parameters, "output_optional_reach_fc").valueAsText
        output_sw_reach_fc_buffer = _get_by_name(parameters, "output_sw_reach_fc_buffer").valueAsText
        output_optional_reach_fc_buffer = _get_by_name(parameters, "output_optional_reach_fc_buffer").valueAsText

        # check for spatial analyst extension
        if arcpy.CheckExtension("Spatial") not in ("Available", "AlreadyInitialized"):
            arcpy.AddError("This tool needs a 'Spatial Analyst' Extension. Please activate it and try again.")
            raise arcpy.ExecuteError

        # check if output layer for optional network routing is provided if the optinoal layer is provided 
        if optional_routing_layer and not output_optional_reach_fc:
            arcpy.AddError("Network reach output parameter for the optional routing layer is missing. Please provide a value and try running the tool again.")
            raise arcpy.ExecuteError

        # ensure the proper gdb exists in the location specified
        standard_gdb = os.path.join(gdb_dir, "Standardized_BMC_Ped_Infra_Assessment.gdb")
        scratch_gdb = os.path.join(gdb_dir, "Scratch_BMC_Ped_Infra_Assessment.gdb")

        # Set environment settings
        arcpy.env.workspace = gdb_dir
        workspaces = arcpy.ListWorkspaces("*", "FileGDB")
        arcpy.env.overwriteOutput = True

        if not bool(set(workspaces) == set([standard_gdb, scratch_gdb])):
            arcpy.AddError("Geodatabase names are not valid. This can result from A) incorrect geodatabase names, B) missing geodatabases, C) too many geodatabases in folder. Please refer to the manual for the required geodatabases.")
            raise arcpy.ExecuteError

        # check if the sidewalk network is in standard gdb
        if not arcpy.Exists(os.path.join(standard_gdb, getLayerName(layerInfo, "Pedestrian Network"))):
            arcpy.AddError("The tool needs 'Pedestrian Network' in the standard geodatabase. Please upload the network and try again")
            raise arcpy.ExecuteError
        
        # create a copy of the destinations layer and buffer it to the given distance
        arcpy.AddMessage("Applying necessary buffers to the destinations layer...")
        tmp_destinations_fc = os.path.join(scratch_gdb, "tmp_destinations_fc")
        copy_fc_with_fields(in_fc=destination_fc, out_fc=tmp_destinations_fc, fields=[], scratch_gdb=scratch_gdb)
        tmp_destinations_buffer = os.path.join(scratch_gdb, "tmp_destinations_buffer")
        arcpy.Buffer_analysis(tmp_destinations_fc, tmp_destinations_buffer, buffer_distance, dissolve_option="ALL")

        # calculate euclidean reach
        arcpy.AddMessage("Calculating euclidean reach...")
        arcpy.Buffer_analysis(tmp_destinations_buffer, output_euclidean_reach_fc, f"{travel_distance} Miles", dissolve_option="ALL")

        # calculate sidewalk network reach
        max_distance = 5280*travel_distance
        arcpy.AddMessage("Calculating sidewalk network reach...")
        tmp_sidewalk_network_fc = os.path.join(scratch_gdb, "tmp_sidewalk_network_fc")
        copy_fc_with_fields(
            in_fc=os.path.join(standard_gdb, getLayerName(layerInfo, "Pedestrian Network")),
            out_fc=tmp_sidewalk_network_fc,
            fields=[],
            where_clause="CateTitle <> 'crosswalk_unmarked'",
            scratch_gdb=scratch_gdb
        )

        tmp_sw_network = os.path.join(scratch_gdb, "tmp_sw_network")
        arcpy.Clip_analysis(tmp_sidewalk_network_fc, output_euclidean_reach_fc, tmp_sw_network)

        # see if any destination buffer features intersect with sidewalk network
        tmp_intersect = os.path.join(scratch_gdb, "tmp_intersect")
        arcpy.Intersect_analysis([tmp_destinations_buffer, tmp_sw_network], tmp_intersect)
        intersect_count = int(arcpy.GetCount_management(tmp_intersect).getOutput(0))
        if intersect_count == 0:
            arcpy.AddWarning(f"None of the destination points are within {buffer_distance} of the sidewalk network. \nSidewalk network travelshed cannot be calculated. Try increasing buffer distance to get this output.")
        else:
            arcpy.CalculateField_management(tmp_sw_network, "cost", "1", "PYTHON3", '', "DOUBLE")

            tmp_cost_raster = os.path.join(scratch_gdb, "tmp_cost_raster")
            arcpy.FeatureToRaster_conversion(tmp_sw_network, "cost", tmp_cost_raster, 10)

            tmp_sw_reach_raster = os.path.join(scratch_gdb, "tmp_sw_reach_raster")
            outCostDist = arcpy.sa.CostDistance(tmp_destinations_buffer, tmp_cost_raster, source_capacity=max_distance)
            tmp_raster = os.path.join(scratch_gdb, "tmp_raster")
            outCostDist.save(tmp_raster)
            output_raster = arcpy.sa.Con(tmp_raster, 1, "", "Value >= 0")
            output_raster.save(tmp_sw_reach_raster)
            tmp_sw_reach_fc = os.path.join(scratch_gdb, "tmp_sw_reach_fc")
            arcpy.RasterToPolygon_conversion(tmp_sw_reach_raster, tmp_sw_reach_fc, "NO_SIMPLIFY", "Value", "SINGLE_OUTER_PART", None)
            # clip the input network to the sw_reach raster cells
            tmp_sw_reach_fc_clip = os.path.join(scratch_gdb, "tmp_sw_reach_fc_clip")
            arcpy.Clip_analysis(tmp_sw_network, tmp_sw_reach_fc, tmp_sw_reach_fc_clip)

            # travelshed output
            arcpy.CopyFeatures_management(tmp_sw_reach_fc_clip, output_sw_reach_fc)
            # generate travelshed as a buffer
            arcpy.Buffer_analysis(tmp_sw_reach_fc_clip, output_sw_reach_fc_buffer, "100 Feet", dissolve_option="ALL")
            # arcpy.RasterToPolyline_conversion(tmp_sw_reach_raster, output_sw_reach_fc, "NODATA")

        # calculate optional network reach
        # if optional_routing_layer and False:
        # check for advanced extension
        if optional_routing_layer and arcpy.CheckProduct("Arcinfo") not in ("Available", "AlreadyInitialized"):
            arcpy.AddWarning("Optional layer routing is only possible with an 'Advanced' license.")
        elif optional_routing_layer and arcpy.CheckProduct("Arcinfo") in ("Available", "AlreadyInitialized"):
            tmp_optional_layer = os.path.join(scratch_gdb, "tmp_optional_layer")
            copy_fc_with_fields(in_fc=optional_routing_layer, out_fc=tmp_optional_layer, fields=[], scratch_gdb=scratch_gdb)
            tmp_optional_layer_clip = os.path.join(scratch_gdb, "tmp_optional_layer_clip")
            arcpy.Clip_analysis(tmp_optional_layer, output_euclidean_reach_fc, tmp_optional_layer_clip)
            arcpy.AddMessage("Calculating network reach for the additional layer provided...")
            if append_to_sw_network:
                # buffer networks by a little to convert into polygon type
                tmp_sw_network_poly = os.path.join(scratch_gdb, "tmp_sw_network_poly")
                arcpy.Buffer_analysis(tmp_sw_network, tmp_sw_network_poly, "0.1 Feet")

                tmp_optional_layer_clip_poly = os.path.join(scratch_gdb, "tmp_optional_layer_clip_poly")
                arcpy.Buffer_analysis(tmp_optional_layer_clip, tmp_optional_layer_clip_poly, "0.1 Feet")

                tmp_merged_fc = os.path.join(scratch_gdb, "tmp_merged_fc")
                arcpy.management.Merge(
                    inputs = [tmp_sw_network_poly, tmp_optional_layer_clip_poly],
                    output=tmp_merged_fc
                )
            else:
                tmp_optional_layer_clip_poly = os.path.join(scratch_gdb, "tmp_optional_layer_clip_poly")
                arcpy.Buffer_analysis(tmp_optional_layer_clip, tmp_optional_layer_clip_poly, "0.1 Feet")
                tmp_merged_fc = tmp_optional_layer_clip_poly

            # see if any destination buffer features intersect with optional network
            tmp_intersect = os.path.join(scratch_gdb, "tmp_intersect")
            arcpy.Intersect_analysis([tmp_destinations_buffer, tmp_merged_fc], tmp_intersect)
            intersect_count = int(arcpy.GetCount_management(tmp_intersect).getOutput(0))
            if intersect_count == 0:
                arcpy.AddWarning(f"None of the destination points are within {buffer_distance} of the optional network provided. \nOptional network travelshed cannot be calculated. Try increasing buffer distance to get this output.")
            else:
                arcpy.management.CalculateField(tmp_merged_fc, "cost", "1", "PYTHON3", '', "DOUBLE")
                tmp_cost_raster = os.path.join(scratch_gdb, "tmp_cost_raster")
                arcpy.PolygonToRaster_conversion(tmp_merged_fc, "cost", tmp_cost_raster, "MAXIMUM_COMBINED_AREA", "cost", 10, "BUILD")

                tmp_optional_layer_reach_raster = os.path.join(scratch_gdb, "tmp_optional_layer_reach_raster")
                outCostDist = arcpy.sa.CostDistance(tmp_destinations_buffer, tmp_cost_raster, source_capacity=max_distance)
                tmp_raster = os.path.join(scratch_gdb, "tmp_raster")
                outCostDist.save(tmp_raster)
                output_raster = arcpy.sa.Con(tmp_raster, 1, "", "Value >= 0")
                output_raster.save(tmp_optional_layer_reach_raster)

                tmp_optional_reach_fc = os.path.join(scratch_gdb, "tmp_optional_reach_fc")
                arcpy.RasterToPolygon_conversion(tmp_optional_layer_reach_raster, tmp_optional_reach_fc, "NO_SIMPLIFY", "Value", "SINGLE_OUTER_PART", None)
                # clip the input network to the optional reach raster cells
                tmp_optional_reach_fc_clip = os.path.join(scratch_gdb, "tmp_optional_reach_fc_clip")
                arcpy.Clip_analysis(tmp_merged_fc, tmp_optional_reach_fc, tmp_optional_reach_fc_clip)

                # travelshed output
                arcpy.CopyFeatures_management(tmp_optional_reach_fc_clip, output_optional_reach_fc)
                # generate travelshed as a buffer
                arcpy.Buffer_analysis(tmp_optional_reach_fc_clip, output_optional_reach_fc_buffer, "100 Feet", dissolve_option="ALL")

            
            # int_outCostDist = arcpy.sa.Int(outCostDist)
            # int_outCostDist.save(tmp_optional_layer_reach_raster)
            # arcpy.RasterToPolyline_conversion(tmp_optional_layer_reach_raster, output_optional_reach_fc, "NODATA")

        return True


def _get_by_name(param_list, name):
    list = [param_list[i].name for i in range(len(param_list))]
    return param_list[list.index(name)]