import arcpy
import pandas as pd
import collections
from collections import defaultdict
import os
from statistics import mean

template_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "templates"))
symbology_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "Symbology"))
spatial_ref = arcpy.SpatialReference(2248)

def is_field_unique(fc, field):
    """
    Checks to see if a field in a feature class is unique
    args: 
        fc    -- feature class
        field -- name of the field to check uniqueness for
    """
    field_count = len(collections.Counter([r[0] for r in arcpy.da.SearchCursor(fc, [field])]))
    feat_count =  int(arcpy.GetCount_management(fc).getOutput(0))

    if feat_count > field_count:
        return False
    else:
        return True

def is_pd_field_unqiue(pandasdf, field):
    return len(pandasdf[field].unique()) == len(pandasdf.index)

def _get_field_type(fc, field):
    return arcpy.ListFields(fc)[field].type

def is_multipart(fc, ws):
    """
    Checks if a feature class has any multipart geometries
    args: 
        fc    -- feature class
        ws    -- workspace
    """
    tmp_fc = arcpy.CreateScratchName(data_type="featureclass", workspace=ws)
    orig_count =  int(arcpy.GetCount_management(fc).getOutput(0))
    arcpy.management.MultipartToSinglepart(fc, tmp_fc)
    updated_count = int(arcpy.GetCount_management(tmp_fc).getOutput(0))
    arcpy.management.Delete(tmp_fc)
    if updated_count != orig_count:
        return True
    else:
        return False

def feature_class_to_df(in_fc, fields, nulls):
    """
    Load feature class data into a Pandas Data Frame
    args
        in_fc  -- Input feature class.
        fields -- Fields for input.
    """
    return pd.DataFrame(
        arcpy.da.FeatureClassToNumPyArray(
            in_table=in_fc,
            field_names=fields,
            skip_nulls=False,
            null_value=nulls
        )
    )

def keep_fields(field_info,column_names):
    """
    Hides all fields in the given field_info except those matching a name
    in column_names
    """
    column_names = list(set(column_names)) # remove duplicates
    for i in range(0,field_info.count):
        if field_info.getFieldName(i) in column_names:
            field_info.setVisible(i,"VISIBLE")
        else:
            field_info.setVisible(i,"HIDDEN")
    return field_info

def copy_fc_with_fields(in_fc, out_fc, fields, scratch_gdb, rename_fields=dict(), where_clause=None):
    """
    Removes extraneous fields from a layer and saves the output fields specified.

    args
    in_fc -- the feature class to standardize
    out_fc -- location to save the result
    fields -- list of fields to save
    rename_fields -- dict of fileds with keys as old names and values as new names
    """
    fields = list(set(fields)) # remove duplicates
    arcpy.management.MakeFeatureLayer(in_fc,"tmp_layer")
    field_info = arcpy.Describe("tmp_layer").fieldInfo
    field_info = keep_fields(field_info,fields)
    arcpy.management.MakeFeatureLayer(in_fc,"tmp_layer",field_info=field_info, where_clause=where_clause)
    tmp_fc = arcpy.CreateScratchName(data_type="featureclass", workspace=scratch_gdb)
    arcpy.management.CopyFeatures("tmp_layer", tmp_fc)
    arcpy.management.Project(tmp_fc, out_fc, spatial_ref)
    arcpy.management.RepairGeometry(out_fc)
    # deleting fields in case they somehow crept through
    fields_list = arcpy.ListFields(out_fc)
    for f in fields_list:
        if f.required is not True and f.name not in fields:
            # arcpy.AddMessage(f"deleting field {f.name} from {out_fc}")
            arcpy.management.DeleteField(out_fc, f.name)
    arcpy.management.Delete(tmp_fc)
    arcpy.management.Delete("tmp_layer")
    if len(rename_fields)>0:
        for k,v in rename_fields.items():
            arcpy.management.AlterField(out_fc,field=k, new_field_name=f"{k}_tmp", new_field_alias=f"{k}_tmp")
            arcpy.management.AlterField(out_fc,field=f"{k}_tmp", new_field_name=v, new_field_alias=v)

# some spatial and table data analysis functions

def agg_list(l, agg):
    "Aggregation method to aggregate items on a list"
    assert agg in ["min", "max", "mean", "sum"], "agg method must be one of ['min', 'max', 'mean', 'sum']"
    clean_list = [i for i in l if i]
    if len(clean_list) == 0:
        # so that the functions return a zero instead of error or None
        clean_list = [0]
    if agg == "min":
        return min(clean_list)
    elif agg == "max":
        return max(clean_list)
    elif agg == "mean":
        return mean(clean_list)
    elif agg == "sum":
        return sum(clean_list)

def transfer_values(to_table, to_field, to_pkey, from_table, from_field, from_pkey, transform_dict=None):
    """
    Transfers field values from one fc to another based on their pkey columns.
    Applies a transformation based on supplied dictionary if provided.
    args:
    to_table -- table to transfer attributes to
    to_field -- field to transfer values to
    to_pkey -- pkey field in the to_table
    from_table -- table to transfer attributes from
    from_field -- field to transfer values from
    from_pkey -- pkey field in the from_table
    transform_dict -- dictionary if needed to transfrom values
    """
    attr_dict = dict()
    cursor = arcpy.da.SearchCursor(from_table, [from_pkey, from_field])
    for row in cursor:
        if transform_dict:
            attr_dict[row[0]] = transform_dict[row[1]]
        else:
            attr_dict[row[0]] = row[1]
    
    cursor = arcpy.da.UpdateCursor(to_table, [to_pkey, to_field])
    for row in cursor:
        row[1] = attr_dict[row[0]]
        cursor.updateRow(row)

def count_within_dist(in_fc, field, input_uid, target_fc, agg, scratch_gdb, search_dist=None, target_field=None):
    """
    Count the number of features in target_fc within a given dist of features in in_fc.
    If target_field is given, it aggergates the target_field values based on agg function.
    """
    tmp_fc_join = os.path.join(scratch_gdb, "tmp_fc_join")
    if search_dist:
        arcpy.SpatialJoin_analysis(
            in_fc,
            target_fc,
            tmp_fc_join,
            join_operation="JOIN_ONE_TO_MANY",
            join_type="KEEP_COMMON",
            match_option="WITHIN_A_DISTANCE",
            search_radius=search_dist
        )
    else:
        arcpy.SpatialJoin_analysis(
            in_fc,
            target_fc,
            tmp_fc_join,
            join_operation="JOIN_ONE_TO_MANY",
            join_type="KEEP_COMMON",
            match_option="INTERSECT"
        )
    # read the join info based on uid
    if not target_field:
        target_field = "Join_Count"
    cursor = arcpy.da.SearchCursor(tmp_fc_join, [input_uid, target_field])
    attr_dict = defaultdict(list)
    for row in cursor:
        attr_dict[row[0]].append(row[1])
    
    # write the count infor back to in_fc
    cursor = arcpy.da.UpdateCursor(in_fc, [input_uid, field])
    for row in cursor:
        row[1] = agg_list(l=attr_dict[row[0]], agg=agg)
        cursor.updateRow(row)

    arcpy.Delete_management(tmp_fc_join)

def calculate_density_by_intersect(input_fc, input_uid, target_fc, target_uid, field_dict, scratch_gdb, area_multiplier):
    """
    Calculates the density of a variable and writes to the input_fc from a corresponding density field in target_fc.
    area_exp must be an expression for calculating area in a given units
    --
    input_fc -- input feature class
    input_uid -- unique id field of input_fc
    target_fc -- feature class to calculate density from
    target_uid -- unique id field of target_fc
    field_dict -- dictionary {k : v} where k = input_fc density field to write to and v = target_fc field. The k and v values should not have an overlap
    """
    # calculate density in the target_layer for each of the fields
    target_feat_area = dict()
    # target_flds = list(field_dict.values())
    # for v in target_flds:
    #     density_dict[v] = defaultdict(int)
    cursor = arcpy.da.SearchCursor(target_fc, [target_uid, "SHAPE@AREA"])
    for row in cursor:
        target_feat_area[row[0]] = row[1]
        # for i in range(2, len(target_flds)+2):
        #     density_dict[target_flds[i-2]][row[0]] = row[i]/row[1]
    
    # calculate the total based on density and overlap area
    tmp_pw_intersect = os.path.join(scratch_gdb, "tmp_pw_intersect")
    arcpy.analysis.PairwiseIntersect([input_fc, target_fc], tmp_pw_intersect, "ALL")
    
    input_feat_area = defaultdict(int)
    cursor = arcpy.da.SearchCursor(tmp_pw_intersect, [input_uid, "SHAPE@AREA"])
    for row in cursor:
        input_feat_area[row[0]] += row[1]
    
    for k, v in field_dict.items():
        attr_dict = defaultdict(int)
        cursor = arcpy.da.SearchCursor(tmp_pw_intersect, [input_uid, target_uid, v, "SHAPE@AREA"])
        for row in cursor:
            attr_dict[row[0]] += (row[2]/target_feat_area[row[1]])*(row[3])
        
        cursor = arcpy.da.UpdateCursor(input_fc, [input_uid, k])
        for row in cursor:
            if input_feat_area[row[0]] == 0:
                row[1] = 0
            else:
                row[1] = (attr_dict[row[0]] / input_feat_area[row[0]]) * area_multiplier
            cursor.updateRow(row)

def define_field(in_layer, field_name):
    """
    defines a new field and checks to ensure field doesn't already exist. If so, then adds 1 to the name. 
    args: 
        - in_layer [fc]: feature class to check for existing field name 
        - candidate field name
    """
    exists = False
    n = 1
    while not exists:
        if field_name in arcpy.ListFields(in_layer):
            field_name = field_name + '_' + n
            n += 1
        else:
            exists = True
    return field_name

def get_attr_from_polygons(in_fc, field, input_uid, target_fc, target_field, agg, scratch_gdb, target_field_dict=None):
    """
    Clips input features with target features and writes the target field values to input field.
    in_fc: input polyline layer.
    field: field to write values to.
    target_fc: polygon layer to transfer attributes from.
    target_field: target field to transfer values from.
    target_field_dict: dict item to use with target_field_values as keys. Values must be numeric.
    agg: aggregation method ("max", "min", "sum", "mean").
    """
    tmp_fc_intersect = os.path.join(scratch_gdb, "tmp_fc_intersect")
    arcpy.analysis.PairwiseIntersect(
        in_features = f"'{in_fc}';'{target_fc}'",
        out_feature_class = tmp_fc_intersect,
        join_attributes = "ALL"
    )
    if target_field_dict:
        attr_field = define_field(tmp_fc_intersect, "tmp_attr")
        arcpy.management.AddField(tmp_fc_intersect, attr_field, "DOUBLE")
        cursor = arcpy.da.UpdateCursor(tmp_fc_intersect, [target_field, attr_field])
        for row in cursor:
            row[1] = target_field_dict[row[0]]
            cursor.updateRow(row)
        del cursor
    else:
        attr_field = target_field
    
    # wrtite attribute values into a dict
    attr_dict = defaultdict(list)
    cursor = arcpy.da.SearchCursor(tmp_fc_intersect, [input_uid, attr_field])
    for row in cursor:
        attr_dict[row[0]].append(row[1])
    del cursor

    # write the attr info back to the original fc
    cursor = arcpy.da.UpdateCursor(in_fc, [input_uid, field])
    for row in cursor:
        row[1] = agg_list(l=attr_dict[row[0]], agg=agg)
        cursor.updateRow(row)
    
    del cursor